import sys
import textwrap
from collections import OrderedDict
from contextlib import contextmanager
import datetime
import time
from artemis.fileman.local_dir import make_file_dir
from artemis.general.should_be_builtins import izip_equal
import numpy as np
from si_prefix import si_format
from six import string_types
from six.moves import xrange, StringIO

__author__ = 'peter'


def arraystr(arr, print_threshold, summary_threshold):
    """
    :param arr: A string summary of an array.
    :param print_threshold:
    :param summary_threshold:
    :return:
    """
    if arr.size<print_threshold:
        return '<{type} with shape={shape}, dtype={dtype}, value={value}, at {id}>'.format(
            type=type(arr).__name__, shape=arr.shape, dtype=arr.dtype, value=str(arr).replace('\n', ','), id=hex(id(arr)))
    elif arr.size<summary_threshold:
        return '<{type} with shape={shape}, dtype={dtype}, in=[{min:.3g}, {max:.3g}], at {id}>'.format(
            type=type(arr).__name__, shape=arr.shape, dtype=arr.dtype, min = arr.min(), max=arr.max(), id=hex(id(arr)))
    else:
        return '<{type} with shape={shape}, dtype={dtype}, at {id}>'.format(
            type=type(arr).__name__, shape=arr.shape, dtype=arr.dtype, id=hex(id(arr)))


def equalize_string_lengths(arr, side = 'left'):
    """
    Equalize the lengths of the string representations of the contents of the array.
    :param arr:
    :return:
    """
    assert side in ('left', 'right')
    strings = [str(x) for x in arr]
    longest = max(len(x) for x in strings)
    if side=='left':
        strings = [string.ljust(longest) for string in strings]
    else:
        strings = [string.rjust(longest) for string in strings]
    return strings


def sensible_str(data, size_limit=4, compact=True):
    """
    Crawl through an data structure and try to make a sensible compact representation of it.
    :param data: Some data structure.
    :param size_limit: The max number of elements in a collection to show.
    :param compact: Remove spaces from output string.
    :return: A one-line string giving a "sensible" overview of what's in the data structure.
    """
    if isinstance(data, np.ndarray):
        if data.size<=size_limit:
            string = 'ndarray('+str(data).replace('\n',',')+')'
        else:
            string = '<{} ndarray>'.format(str(data.shape).replace(' ', ''))
    elif isinstance(data, (list, tuple)):
        if len(data)>size_limit:
            string = '<len{}-{}>'.format(len(data), data.__class__.__name__)
        else:
            open, close = '[]' if isinstance(data, list) else '()'
            string = open +', '.join(sensible_str(x) for x in data[:size_limit]) + close
    elif isinstance(data, dict):
        if len(data)>size_limit:
            string = '<len{}-{}>'.format(len(data), data.__class__.__name__)
        else:
            open, close = ('OrderedDict([', '])') if isinstance(data, OrderedDict) else '{}'
            string = open+', '.join('{}:{}'.format(sensible_str(k), sensible_str(v)) for i, (k, v) in zip(range(size_limit), data.items())) + close
    else:
        string = str(data).replace('\n', '\\n')

    if compact:
        string = string.replace(' ', '')
    return string


@contextmanager
def hold_numpy_printoptions(**kwargs):
    """
    Temporarily set the numpy print options.
    See https://docs.scipy.org/doc/numpy/reference/generated/numpy.set_printoptions.html
    :param kwargs: Print options (see link)
    """
    opts = np.get_printoptions()
    np.set_printoptions(**kwargs)
    yield
    np.set_printoptions(**opts)


def str_with_arrayopts(obj, float_format='.3g', threshold=8, **kwargs):
    """
    Print
    :param obj:
    :param float_format:
    :param threshold:
    :param kwargs:
    :return:
    """
    with hold_numpy_printoptions(formatter = {'float': lambda x: '{{:{}}}'.format(float_format).format(x)}, threshold=threshold, **kwargs):
        return str(obj)


def deepstr(obj, memo=None, array_print_threhold = 8, array_summary_threshold=10000, max_expansion = None, indent ='  ', float_format = ''):
    """
    A recursive, readable print of a data structure.
    """
    if memo is None:
        memo = set()

    if id(obj) in memo:
        return "<{type} at {loc}> already listed".format(type=type(obj).__name__, loc=hex(id(obj)))
    memo.add(id(obj))

    if isinstance(obj, np.ndarray):
        string_desc = arraystr(obj, print_threshold=array_print_threhold, summary_threshold=array_summary_threshold)
    elif isinstance(obj, (list, tuple, set, dict)):
        kwargs = dict(memo=memo, array_print_threhold=array_print_threhold, max_expansion=max_expansion, array_summary_threshold=array_summary_threshold, indent=indent, float_format=float_format)

        if isinstance(obj, (list, tuple)):
            keys, values = [str(i) for i in xrange(len(obj))], obj
        elif isinstance(obj, dict):
            keys, values = obj.keys(), obj.values()
        elif isinstance(obj, set):
            keys, values = ['- ']*len(obj), obj
        else:
            raise Exception('Should never be here')
        max_indent = max(len(str(k)) for k in keys) if len(keys)>0 else 0
        if max_expansion is not None and len(keys)>max_expansion:
            elements = ['{k}: {v}'.format(k=k, v=' '*(max_indent-len(str(k))) + indent_string(deepstr(v, **kwargs), indent=' '*max_indent, include_first=False)) for k, v in izip_equal(keys[:max_expansion-1]+[keys[-1]], values[:max_expansion-1]+[values[-1]])]
            elements.insert(-1, '... Skipping {} of {} elements ...'.format(len(keys)-len(elements), len(keys)))
        else:
            elements = ['{k}: {v}'.format(k=k, v=' '*(max_indent-len(str(k))) + indent_string(deepstr(v, **kwargs), indent=' '*max_indent, include_first=False)) for k, v in izip_equal(keys, values)]
        string_desc = '<{type} at {id}>\n'.format(type = type(obj).__name__, id=hex(id(obj))) + indent_string('\n'.join(elements), indent=indent)
        return string_desc
    elif isinstance(obj, float):
        string_desc = '{{:{}}}'.format(float_format).format(obj)
    else:
        string_desc = str(obj)
    return string_desc


_ORIGINAL_STDOUT = sys.stdout
_ORIGINAL_STDERR = sys.stderr


class CaptureStdOut(object):
    """
    An logger that both prints to stdout and writes to file.
    """

    def __init__(self, log_file_path = None, print_to_console = True, prefix = None):
        """
        :param log_file_path: The path to save the records, or None if you just want to keep it in memory
        :param print_to_console:
        """
        self._print_to_console = print_to_console
        if log_file_path is not None:
            # self._log_file_path = os.path.join(base_dir, log_file_path.replace('%T', now))
            make_file_dir(log_file_path)
            self.log = open(log_file_path, 'w')
        else:
            self.log = StringIO()
        self._log_file_path = log_file_path
        self.old_stdout = _ORIGINAL_STDOUT
        self.prefix = None if prefix is None else prefix

    def __enter__(self):

        self.old_stdout = sys.stdout
        self.old_stderr = sys.stderr

        sys.stdout = self
        sys.stderr = self
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        sys.stdout.flush()
        sys.stderr.flush()
        sys.stdout = self.old_stdout
        sys.stderr = self.old_stderr
        self.close()

    def get_log_file_path(self):
        assert self._log_file_path is not None, "You never specified a path when you created this logger, so don't come back and ask for one now"
        return self._log_file_path

    def write(self, message):
        if self._print_to_console:
            self.old_stdout.write(message if self.prefix is None or message=='\n' else self.prefix+message)
        self.log.write(message)
        self.log.flush()

    def close(self):
        if self._log_file_path is not None:
            self.log.close()

    def read(self):
        if self._log_file_path is None:
            return self.log.getvalue()
        else:
            with open(self._log_file_path) as f:
                txt = f.read()
            return txt

    def __getattr__(self, item):
        return getattr(self.old_stdout, item)


def indent_string(str, indent = '  ', include_first = True, include_last = False):
    base = str.replace('\n', '\n'+indent)
    if include_first:
        base = indent + base
    if not include_last and base.endswith('\n'+indent):
        base = base[:-len(indent)]
    return base


class IndentPrint(object):
    """
    Indent all print statements
    """

    def __init__(self, block_header=None, spacing = 4, show_line = False, show_end = False):
        self.indent = '|'+' '*(spacing-1) if show_line else ' '*spacing
        self.show_end = show_end
        self.block_header = block_header

    def __enter__(self):
        self.old_stdout = sys.stdout
        if self.block_header is not None:
            print(self.block_header)
        sys.stdout = self

    def flush(self):
        self.old_stdout.flush()

    def write(self, message):
        if message=='\n':
            new_message = '\n'
        else:
            new_message = indent_string(message, self.indent)
        self.old_stdout.write(new_message)

    def close(self):
        self.old_stdout.close()

    def __exit__(self, exc_type, exc_val, exc_tb):
        sys.stdout = self.old_stdout
        if self.show_end:
            print('```')


class DocumentWrapper(textwrap.TextWrapper):

    def wrap(self, text):
        split_text = text.split('\n')
        lines = [line for para in split_text for line in textwrap.TextWrapper.wrap(self, para)]
        return lines


def side_by_side(multiline_strings, gap=4, max_linewidth=None):
    """
    Return a string that displays two multiline strings side-by-side.
    :param multiline_strings: A list of multi-line strings (ie strings with newlines)
    :param gap: The gap (in spaces) to make between side-by-side display of the strings
    :param max_linewidth: Maximum width of the strings
    :return: A single string which shows all the above strings side-by-side.
    """
    if max_linewidth is not None:
        w = DocumentWrapper(width=max_linewidth, replace_whitespace=False)
        lineses = [w.wrap(mlstring) for mlstring in multiline_strings]
    else:
        lineses = [s.split('\n') for s in multiline_strings]

    if all(len(lines)==0 for lines in lineses):  # All strings are empty
        return ''

    longests = [max(len(line) for line in lines) if len(lines)>0 else 0 for lines in lineses]

    spacer = ' '*gap
    new_lines = []
    for i in xrange(max(len(lines) for lines in lineses)):
        line = [lines[i] if i<len(lines) else '' for lines in lineses]
        final_line = [line + ' '*(max_length-len(line)) for line, max_length in zip(line, longests)]
        new_lines.append(spacer.join(final_line))
    return '\n'.join(new_lines)


def truncate_string(string, truncation, message = ''):
    """
    Truncate a string to a given length.  Optionally add a message at the end
    explaining the truncation
    :param string: A string
    :param truncation: An int
    :param message: A message, e.g. '...<truncated>'
    :return: A new string no longer than truncation
    """
    if truncation is None:
        return string
    assert isinstance(truncation, int)
    if len(string)>truncation:
        return string[:truncation-len(message)]+message
    else:
        return string


def surround_with_header(string, width, char='-'):
    """
    Surround the string by a header.  The result has length min(len(string)+2, width)
    :param string: A string
    :param width: Width of the entire header
    :param char: Character to repeat
    :return: A header, whose length will be
    """
    left = (width-len(string)-1)//2
    right = (width-len(string)-2)//2
    return char*left+' '+string+' '+char*right


def section_with_header(header, content, width=50, top_char=None, header_char='-', bottom_char=None):

    string = '' if top_char is None else top_char*width+'\n'
    string += surround_with_header(header, width=width, char=header_char) + '\n'+content+'\n'
    if bottom_char is not None:
        string += bottom_char*width
    return string


@contextmanager
def assert_things_are_printed(things, min_len=None):
    """
    Make sure that the things in theings are preinted.
    :param things:
    :param min_len:
    :return:
    """

    if isinstance(things, string_types):
        things = [things]

    with CaptureStdOut() as cap:
        yield

    printed_text = cap.read()

    if min_len is not None:
        assert len(printed_text) >= min_len, 'Printed text length {} was under the minimum length of {}'.format(len(printed_text), min_len)

    for thing in things:
        assert thing in printed_text, '"{}" was not printed'.format(thing)


_seconds_in_day = 60*60*24


def format_duration(seconds):
    '''
    Formats a float interpreted as seconds as a sensible time duration
    :param seconds:
    :return:
    '''
    if seconds < 60:
        return si_format(seconds, precision=1, format_str='{value}{prefix}s')
    elif seconds<_seconds_in_day:
        res = str(datetime.timedelta(seconds=seconds))
        if len(res.split(".")) > 1:
            return ".".join(res.split(".")[:-1])
        else:
            return res
    else:
        days = seconds//_seconds_in_day
        return '{:d}d,{}'.format(days, format_duration(seconds % _seconds_in_day))


def format_time_stamp(time_stamp):
    if isinstance(time_stamp,str):
        return time_stamp
    else:
        if isinstance(time_stamp,float):
            time_stamp = datetime.datetime.utcfromtimestamp(time_stamp)
        else:
            assert isinstance(time_stamp,datetime.datetime), "Time Stamp not understood"
        if time_stamp.year != time.gmtime(time.time()).tm_year:
            format = "%b %d %Y, %H:%M:%S"
        else:
            format = "%b %d, %H:%M:%S"
        return time_stamp.strftime(format)





import inspect
from collections import OrderedDict
from functools import partial
import collections
from artemis.general.should_be_builtins import separate_common_items


def get_partial_chain(f):
    """
    Given a partial function f, the chain of functions from which it is derived.
    eg:
        def f(a, b):
            return a+b
        g = partial(f, b=3)
        h = partial(g, a=2)
        assert [f, g, h] == get_partial_chain(h)

    WARNING: Python 3 automatically collapses partials - so in Python 3 this chain will never be longer than 2!

    :param f: A function, possibly a partial
    :return: A list of functions, starting with the root
    """
    return get_partial_chain(f.func) + [f] if isinstance(f, partial) else [f]


def get_partial_root(f):
    return get_partial_chain(f)[0]


def infer_function_and_derived_arg_values(f):
    """
    Given a function f, which may be a partial version of some other function, going down to some root, standard python
    function, get the full set of arguments that this final function will end up being called with.  This function will
    raise an AssertionError if not arguments are defined by the partial chain.

    :param f: A function, or partial function
    :return: root_func, kwargs     ... where:
        root_func is the root function.
        kwargs is An OrderedDict(arg_name->arg_value)
    """
    partial_chain = get_partial_chain(f)
    root, partials = partial_chain[0], partial_chain[1:]
    assert all(len(pf.args)==0 for pf in partials), "We don't handle unnamed arguments for now.  Add this functionality if necessary"
    overrides = dict((argname, argval) for pf in partials for argname, argval in pf.keywords.items())  # Note that later updates on the same args go into the dict
    full_arg_list = infer_arg_values(root, **overrides)
    return root, full_arg_list


def infer_derived_arg_values(f):
    """
    Given a function f, which may be a partial version of some other function, going down to some root, standard python
    function, get the full set of arguments that this final function will end up being called with.  This function will
    raise an AssertionError if not arguments are defined by the partial chain.
    e.g.

        def f(a, b=1):
            return a+b
        g = partial(f, a=2)
        h = partial(g, b=3)
        assert f(**get_derived_function_args(g)) == g()
        assert f(**get_derived_function_args(h)) == h()

    :param f: A function, or partial function
    :return: An OrderedDict(arg_name->arg_value)
    """
    _, full_arg_list = infer_function_and_derived_arg_values(f)
    return full_arg_list


def infer_arg_values(f, *args, **kwargs):
    """
    :param f: Get the full list of arguments to a function, or throw an error if the function cannot be called by the
        given arguments.
    :param args: A list of args
    :param kwargs: A dict of keyword args
    :return: An OrderedDict(arg_name->arg_value)
    """
    all_arg_names, varargs_name, kwargs_name, defaults = inspect.getargspec(f)
    assert varargs_name is None, "This function doesn't work with unnamed args"
    default_args = {k: v for k, v in zip(all_arg_names[len(all_arg_names)-(len(defaults) if defaults is not None else 0):], defaults if defaults is not None else [])}
    args_with_values = set(all_arg_names[:len(args)]+list(default_args.keys())+list(kwargs.keys()))
    assert set(all_arg_names).issubset(args_with_values), "Arguments {} require values but are not given any.  ".format(tuple(set(all_arg_names).difference(args_with_values)))
    assert len(args) <= len(all_arg_names), "You provided {} arguments, but the function only takes {}".format(len(args), len(all_arg_names))
    full_args = tuple(
        list(zip(all_arg_names, args))  # Handle unnamed args f(1, 2)
        + [(name, kwargs[name] if name in kwargs else default_args[name]) for name in all_arg_names[len(args):]]  # Handle named keyworkd args f(a=1, b=2)
        + [(name, kwargs[name]) for name in kwargs if name not in all_arg_names[len(args):]]  # Need to handle case if f takes **kwargs
        )
    duplicates = tuple(item for item, count in collections.Counter([a for a, _ in full_args]).items() if count > 1)
    assert len(duplicates)==0, 'Arguments {} have been defined multiple times: {}'.format(duplicates, full_args)

    common_args, (different_args, different_given_args) = separate_common_items([tuple(all_arg_names), tuple(n for n, _ in full_args)])
    if kwargs_name is None:  # There is no **kwargs
        assert len(different_given_args)==0, "Function {} was given args {} but didn't ask for them".format(f, different_given_args)
    assert len(different_args)==0, "Function {} needs values for args {} but didn't get them".format(f, different_args)
    return OrderedDict(full_args)



from artemis.plotting.data_conversion import put_data_in_grid
from matplotlib import pyplot as plt
import numpy as np


def select_image(images, selection_callback):
    """
    Givel an array of images, show them all and allow seection by clicking.
    :param images:
    :param selection_callback:
    :return:
    """

    data = put_data_in_grid(images, is_color_data=True)

    plt.figure(figsize=(12, 12))
    plt.imshow(data)
    plt.gca().tick_params(labelbottom = 'off')
    plt.gca().tick_params(labelleft = 'off')

    # dbplot(first_ims, 'First BBox Images')
    def callback(event):
        n_cols = int(np.ceil(np.sqrt(len(images))))
        n_rows = int(np.ceil(len(images)/n_cols))
        ax_size_x = plt.gca().get_xlim()[1]
        ax_size_y = plt.gca().get_ylim()[0]

        frac_x = event.xdata/ax_size_x
        frac_y = event.ydata/ax_size_y
        print (frac_x, frac_y)

        col_ix = int(n_cols*frac_x)
        row_ix = int((n_rows+1)*frac_y)

        ix = row_ix*n_cols + col_ix

        selection_callback(ix)

    plt.gcf().canvas.callbacks.connect('button_press_event', callback)
    plt.show()


def select_pixel(image, selection_callback):

    # plt.figure()
    plt.imshow(image)

    def callback(event):
        selection_callback((int(event.ydata), int(event.xdata)))

    plt.gcf().canvas.callbacks.connect('button_press_event', callback)
    plt.show()



import shlex
import subprocess
import sys
import threading

from artemis.remote.utils import get_local_ips, get_ssh_connection




def execute_command(ip_address, command, blocking=True):
    '''
    This method spawns a child-process (either locally or remote, depending on the ip_address). Then it executes the given command and handles communication.
    If blocking is True, then this call does not return before the child process returns. It then prints stdout to console, followed by stderr.
    If blocking is False, then this call returns immediately and asynchronously forwards stdout and stderr to the console in separate threads.
    If ip_address is local, then the command will be split using shlex.split() and must be formatted accordingly. The subprocess call is executed with shell=False
    for all the right reasons.
    :param ip_address:
    :param command: String. command to execute
    :param blocking:
    :return:
    '''
    #TODO: Reformat this to work with the Nanny. Let it manage the communication
    if ip_address in get_local_ips():
        sub = subprocess.Popen(shlex.split(command),stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=False)
        stdout_pipe = sub.stdout
        stderr_pipe = sub.stderr

    else:
        ssh_conn = get_ssh_connection(ip_address)
        transport = ssh_conn.get_transport()
        channel = transport.open_session()
        channel.exec_command(command)#
        bufsize=-1
        stdout_pipe = channel.makefile('r', bufsize)
        stderr_pipe = channel.makefile_stderr('r',bufsize)

    #stdout
    t1 = ParamikoPrintThread(source_pipe=stdout_pipe, target_pipe=sys.stdout)
    t1.start()
    # stderr
    t2 = ParamikoPrintThread(source_pipe=stderr_pipe, target_pipe=sys.stderr)
    t2.start()
    if blocking:
        t1.join()
        t2.join()

class ParamikoPrintThread(threading.Thread):
    def __init__(self, source_pipe, target_pipe, stopping_criterium=None, stop_event=None, prefix=""):
        '''
        :param source_pipe: The ssh pipe from which to forward communications
        :param target_pipe: Either stdout or stderr. This determines if stderr or stdout is read from the ssh channel.
        :param stopping_criterium: function which takes in a line from source_pipe and evaluates to boolean.:
        :param prefix: A prefix that is attached to every printed line.
        :return:
        '''
        self.source_pipe = source_pipe
        self.prefix = prefix
        self.target_pipe = target_pipe
        self.stopping_criterium = stopping_criterium
        super(ParamikoPrintThread, self).__init__()

    def run(self, ):
        with self.source_pipe:
            # Terminates when source pipe runs dry.
            for line in iter(self.source_pipe.readline, b''):
                self.target_pipe.write("%s%s" % (self.prefix, line))
                self.target_pipe.flush()
                if self.stopping_criterium is not None and self.stopping_criterium(line):
                    break


from setuptools import setup, find_packages

setup(
    name='artemis-ml',
    author='Peter & Matthias',
    author_email='poconn4@gmail.com',
    url='https://github.com/quva-lab/artemis',
    long_description='Artemis aims to get rid of all the boring, bureaucratic coding (plotting, file management, etc) involved in machine learning projects, so you can get to the good stuff quickly.',
    install_requires=['numpy', 'scipy', 'matplotlib', 'pytest', 'pillow', 'tabulate', 'si-prefix', 'enum34'],
    extras_require = {
        'remote_plotting': ["paramiko", "netifaces"]
        },
    version='2.0.0',
    packages=find_packages(),
    scripts=[])



import re
from collections import OrderedDict

from tabulate import tabulate
from artemis.experiments.experiment_management import load_lastest_experiment_results
from artemis.experiments.experiment_record import NoSavedResultError, ExpInfoFields, ExperimentRecord, \
    load_experiment_record, is_matplotlib_imported
from artemis.experiments.experiments import is_experiment_loadable, get_global_experiment_library
from artemis.general.display import deepstr, truncate_string, hold_numpy_printoptions, side_by_side, CaptureStdOut, \
    surround_with_header, section_with_header
from artemis.general.nested_structures import flatten_struct
from artemis.general.should_be_builtins import separate_common_items, all_equal, bad_value, izip_equal
from artemis.general.tables import build_table
from six import string_types


def get_record_result_string(record, func='deep', truncate_to = None, array_print_threshold=8, array_float_format='.3g', oneline=False):
    """
    Get a string representing the result of the experiment.
    :param record:
    :param func:
    :return:
    """
    with hold_numpy_printoptions(threshold = array_print_threshold, formatter={'float': lambda x: '{{:{}}}'.format(array_float_format).format(x)}):
        if isinstance(func, string_types):
            func = {
                'deep': deepstr,
                'str': str,
                }[func]
        else:
            assert callable(func), 'func must be callable'
        try:
            result = record.get_result()
        except NoSavedResultError:
            return '<No result has been saved>'
        string = func(result)
        if not isinstance(string, string_types):
            string = str(string)

    if truncate_to is not None:
        string = truncate_string(string, truncation=truncate_to, message = '...<truncated>')
    if oneline:
        string = string.replace('\n', ', ')
    return string


def _get_record_info_section(record, header_width):
    return section_with_header('Info', record.info.get_text(), width=header_width)


def _get_record_log_section(record, truncate_logs = None, header_width=64):
    log = record.get_log()
    if truncate_logs is not None and len(log)>truncate_logs:
        log = log[:truncate_logs-100] + '\n\n ... LOG TRUNCATED TO {} CHARACTERS ... \n\n'.format(truncate_logs) + log[-100:]
    return section_with_header('Logs', log, width=header_width)


def _get_record_error_trace_section(record, header_width):
    error_trace = record.get_error_trace()
    if error_trace is None:
        return ''
    else:
        return section_with_header('Error Trace', record.get_error_trace(), width=header_width)


def _get_result_section(record, truncate_result, show_result, header_width):
    assert show_result in (False, 'full', 'deep')
    result_str = get_record_result_string(record, truncate_to=truncate_result, func=show_result)
    return section_with_header('Result', result_str, width=header_width)


def get_record_full_string(record, show_info = True, show_logs = True, truncate_logs = None, show_result ='deep',
        show_exceptions=True, truncate_result = None, include_bottom_border = True, header_width=64, return_list = False):
    """
    Get a human-readable string containing info about the experiment record.

    :param show_info: Show info about experiment (name, id, runtime, etc)
    :param show_logs: Show logs (True, False, or an integer character count to truncate logs at)
    :param show_result: Show the result.  Options for display are:
        'short': Print a one-liner (note: sometimes prints multiple lines)
        'long': Directly call str
        'deep': Use the deepstr function for a compact nested printout.
    :return: A string to print.
    """

    parts = [surround_with_header(record.get_id(), width=header_width, char='=')]

    if show_info:
        parts.append(section_with_header('Info', record.info.get_text(), width=header_width))

    if show_logs:
        log = record.get_log()
        if truncate_logs is not None and len(log)>truncate_logs:
            log = log[:truncate_logs-100] + '\n\n ... LOG TRUNCATED TO {} CHARACTERS ... \n\n'.format(truncate_logs) + log[-100:]
        # return section_with_header('Logs', log, width=header_width)
        parts.append(section_with_header('Logs', log, width=header_width))

    if show_exceptions:
        error_trace = record.get_error_trace()
        error_trace_text = '' if error_trace is None else section_with_header('Error Trace', record.get_error_trace(), width=header_width)
        parts.append(error_trace_text)

    if show_result:
        assert show_result in (False, 'full', 'deep')
        result_str = get_record_result_string(record, truncate_to=truncate_result, func=show_result)
        parts.append(section_with_header('Result', result_str, width=header_width))

    if return_list:
        return parts
    else:
        return '\n'.join(parts)


def get_record_invalid_arg_string(record, recursive=True, ignore_valid_keys=(), note_version = 'full'):
    """
    Return a string identifying ig the arguments for this experiment are still valid.
    :return:
    """
    assert note_version in ('full', 'short')
    experiment_id = record.get_experiment_id()
    if is_experiment_loadable(experiment_id):
        if record.info.has_field(ExpInfoFields.ARGS):
            last_run_args = OrderedDict([(k,v) for k,v in record.get_args().items() if k not in ignore_valid_keys])
            current_args = OrderedDict([(k,v) for k,v in record.get_experiment().get_args().items() if k not in ignore_valid_keys])
            if recursive:
                last_run_args = OrderedDict(flatten_struct(last_run_args, first_dict_is_namespace=True))
                last_run_args = OrderedDict([(k, v) for k, v in last_run_args.items() if k not in ignore_valid_keys])
                current_args = OrderedDict(flatten_struct(current_args, first_dict_is_namespace=True))
                current_args = OrderedDict([(k, v) for k, v in current_args.items() if k not in ignore_valid_keys])

            validity = record.args_valid(last_run_args=last_run_args, current_args=current_args)
            if validity is False:
                last_arg_str, this_arg_str = [['{}:{}'.format(k, v) for k, v in argdict.items()] for argdict in (last_run_args, current_args)]
                common, (old_args, new_args) = separate_common_items([last_arg_str, this_arg_str])
                if len(old_args)+len(new_args)==0:
                    raise Exception('Error displaying different args.  Bug Peter.')
                changestr = "{{{}}}->{{{}}}".format(','.join(old_args), ','.join(new_args))
                notes = ("Change: " if note_version=='full' else "") + changestr
            elif validity is None:
                notes = "Cannot Determine: Unhashable Args" if note_version=='full' else '<Unhashable Args>'
            else:
                notes = "<No Change>"
        else:
            notes = "Cannot Determine: Inconsistent Experiment Record" if note_version == 'full' else '<Inconsistent Record>'
    else:
        notes = "Cannot Determine: Experiment Not Imported" if note_version=='full' else '<Not Imported>'
    return notes


def get_oneline_result_string(record, truncate_to=None, array_float_format='.3g', array_print_threshold=8):
    """
    Get a string that describes the result of the record in one line.  This can optionally be specified by
    experiment.one_liner_function.

    :param record: An ExperimentRecord.
    :param truncate_to:
    :param array_float_format:
    :param array_print_threshold:
    :return: A string with no newlines briefly describing the result of the record.
    """
    if isinstance(record, string_types):
        record = load_experiment_record(record)
    if not is_experiment_loadable(record.get_experiment_id()):
        one_liner_function=str
    else:
        one_liner_function = record.get_experiment().one_liner_function
        if one_liner_function is None:
            one_liner_function = str
    return get_record_result_string(record, func=one_liner_function, truncate_to=truncate_to, array_print_threshold=array_print_threshold,
        array_float_format=array_float_format, oneline=True)


def print_experiment_record_argtable(records):
    """
    Print a table comparing experiment arguments and their results.
    """
    funtion_names = [record.info.get_field(ExpInfoFields.FUNCTION) for record in records]
    args = [record.info.get_field(ExpInfoFields.ARGS) for record in records]
    results = [record.get_result(err_if_none=False) for record in records]

    common_args, different_args = separate_common_items(args)

    record_ids = [record.get_id() for record in records]

    def lookup_fcn(record_id, column):
        index = record_ids.index(record_id)
        if column=='Function':
            return funtion_names[index]
        elif column=='Run Time':
            return records[index].info.get_field_text(ExpInfoFields.RUNTIME)
        elif column=='Common Args':
            return ', '.join('{}={}'.format(k, v) for k, v in common_args)
        elif column=='Different Args':
            return ', '.join('{}={}'.format(k, v) for k, v in different_args[index])
        elif column=='Result':
            return get_oneline_result_string(records[index])
        else:
            bad_value(column)

    rows = build_table(lookup_fcn,
        row_categories=record_ids,
        column_categories=['Function', 'Run Time', 'Common Args', 'Different Args', 'Result'],
        prettify_labels=False
        )

    print(tabulate(rows))


def show_record(record, show_logs=True, truncate_logs=None, truncate_result=10000, header_width=100, show_result ='deep', hang=True):
    """
    Show the results of an experiment record.
    :param record:
    :param show_logs:
    :param truncate_logs:
    :param truncate_result:
    :param header_width:
    :param show_result:
    :param hang:
    :return:
    """
    string = get_record_full_string(record, show_logs=show_logs, show_result=show_result, truncate_logs=truncate_logs,
        truncate_result=truncate_result, header_width=header_width, include_bottom_border=False)

    has_matplotlib_figures = any(loc.endswith('.pkl') for loc in record.get_figure_locs())
    if has_matplotlib_figures:
        from matplotlib import pyplot as plt
        from artemis.plotting.saving_plots import interactive_matplotlib_context
        record.show_figures(hang=hang)
    print(string)


def show_multiple_records(records, func = None):

    if func is None:
        func = lambda rec: rec.get_experiment().show(rec)

    if is_matplotlib_imported():
        from artemis.plotting.manage_plotting import delay_show
        with delay_show():
            for rec in records:
                func(rec)
    else:
        for rec in records:
            func(rec)


def compare_experiment_records(records, parallel_text=None, show_logs=True, truncate_logs=None,
        truncate_result=10000, header_width=100, max_linewidth=128, show_result ='deep'):
    """
    Show the console logs, figures, and results of a collection of experiments.

    :param records:
    :param parallel_text:
    :param hang_notice:
    :return:
    """
    if isinstance(records, ExperimentRecord):
        records = [records]
    if parallel_text is None:
        parallel_text = len(records)>1
    if len(records)==0:
        print('... No records to show ...')
        return
    else:
        records_sections = [get_record_full_string(rec, show_logs=show_logs, show_result=show_result, truncate_logs=truncate_logs,
                    truncate_result=truncate_result, header_width=header_width, include_bottom_border=False, return_list=True) for rec in records]

    if parallel_text:
        full_string = '\n'.join(side_by_side(records_section, max_linewidth=max_linewidth) for records_section in zip(*records_sections))
    else:
        full_string = '\n'.join('\n'.join(record_sections) for record_sections in records_sections)

    print(full_string)

    has_matplotlib_figures = any(loc.endswith('.pkl') for rec in records for loc in rec.get_figure_locs())
    if has_matplotlib_figures:
        from artemis.plotting.manage_plotting import delay_show
        with delay_show():
            for rec in records:
                rec.show_figures()

    return has_matplotlib_figures


def find_experiment(*search_terms):
    """
    Find an experiment.  Invoke
    :param search_term: A term that will be used to search for an experiment.
    :return:
    """
    global_lib = get_global_experiment_library()
    found_experiments = OrderedDict((name, ex) for name, ex in global_lib.items() if all(re.search(term, name) for term in search_terms))
    if len(found_experiments)==0:
        raise Exception("None of the {} experiments matched the search: '{}'".format(len(global_lib), search_terms))
    elif len(found_experiments)>1:
        raise Exception("More than one experiment matched the search '{}', you need to be more specific.  Found: {}".format(search_terms, found_experiments.keys()))
    else:
        return found_experiments.values()[0]


def make_record_comparison_table(records, args_to_show=None, results_extractor = None, print_table = False):
    """
    Make a table comparing the arguments and results of different experiment records.  You can use the output
    of this function with the tabulate package to make a nice readable table.

    :param records: A list of records whose results to compare
    :param args_to_show: A list of arguments to show.  If none, it will just show all arguments
        that differ between experiments.
    :param results_extractor: A dict<str->callable> where the callables take the result of the
        experiment as an argument and return an entry in the table.
    :param print_table: Optionally, import tabulate and print the table here and now.
    :return: headers, rows
        headers is a list of of headers for the top of the table
        rows is a list of lists filling in the information.

    example usage:

        headers, rows = make_record_comparison_table(
            record_ids = [experiment_id_to_latest_record_id(eid) for eid in [
                'demo_fast_weight_mlp.multilayer_baseline.1epoch.version=mlp',
                'demo_fast_weight_mlp.multilayer_baseline.1epoch.full-gd.n_steps=1',
                'demo_fast_weight_mlp.multilayer_baseline.1epoch.full-gd.n_steps=20',
                ]],
            results_extractor={
                'Test': lambda result: result.get_best('test').score.get_score('test'),
                'Train': lambda result: result.get_best('test').score.get_score('train'),
                }
             )
        import tabulate
        print tabulate.tabulate(rows, headers=headers, tablefmt=tablefmt)
    """

    args = [rec.info.get_field(ExpInfoFields.ARGS) for rec in records]
    if args_to_show is None:
        common, separate = separate_common_items(args)
        args_to_show = [k for k, v in separate[0]]

    if results_extractor is None:
        results_extractor = {'Result': str}
    elif callable(results_extractor):
        results_extractor = {'Result': results_extractor}
    else:
        assert isinstance(results_extractor, dict)

    headers = args_to_show + results_extractor.keys()

    rows = []
    for record, record_args in izip_equal(records, args):
        arg_dict = dict(record_args)
        args_vals = [arg_dict[k] for k in args_to_show]
        results = record.get_result()
        rows.append(args_vals+[f(results) for f in results_extractor.values()])

    if print_table:
        import tabulate
        print(tabulate.tabulate(rows, headers=headers, tablefmt='simple'))
    return headers, rows


def separate_common_args(records, return_dict = False):
    """

    :param records: A List of records
    :param return_dict: Return the different args as a dict<ExperimentRecord: args>
    :return: (common, different)
        Where common is an OrderedDict of common args
        different is a list (the same lengths of records) of OrderedDicts containing args that are not the same in all records.
    """
    common, argdiff = separate_common_items([list(rec.get_args().items()) for rec in records])
    if return_dict:
        argdiff = {rec.get_id(): args for rec, args in zip(records, argdiff)}
    return common, argdiff



import numpy as np
import pytest
from artemis.general.nondeterminism_hunting import delete_vars, assert_variable_matches_between_runs, variable_matches_between_runs, \
    reset_variable_tracker


def _runs_are_the_same(var_gen_1, var_gen_2, use_assert = False):

    delete_vars(['_test_random_var_32r5477w32'])
    for run, gen in [(0, var_gen_1), (1, var_gen_2)]:
        reset_variable_tracker()
        for v in gen:
            if use_assert:
                assert_variable_matches_between_runs(v, '_test_random_var_32r5477w32')
            else:
                its_a_match=variable_matches_between_runs(v, '_test_random_var_32r5477w32')
                if run==0:
                    assert its_a_match is None
                else:
                    if not its_a_match:
                        return False
    return True


def test_variable_matches_between_runs():

    rng1 = np.random.RandomState(1234)
    gen1 = (rng1.randn(3, 4) for _ in range(5))
    rng2 = np.random.RandomState(1234)
    gen2 = (rng2.randn(3, 4) for _ in range(5))
    assert _runs_are_the_same(gen1, gen2)

    rng = np.random.RandomState(1234)
    gen1 = (rng.randn(3, 4) for _ in range(5))
    gen2 = (rng.randn(3, 4) for _ in range(5))
    assert not _runs_are_the_same(gen1, gen2)

    gen1 = (i for i in range(5))
    gen2 = (i for i in range(5))
    assert _runs_are_the_same(gen1, gen2)

    gen1 = (i for i in range(5))
    gen2 = (i if i<4 else 7 for i in range(5))
    assert not _runs_are_the_same(gen1, gen2)


def test_assert_variable_matches_between_runs():

    rng1 = np.random.RandomState(1234)
    gen1 = (rng1.randn(3, 4) for _ in range(5))
    rng2 = np.random.RandomState(1234)
    gen2 = (rng2.randn(3, 4) for _ in range(5))
    _runs_are_the_same(gen1, gen2, use_assert=True)

    rng = np.random.RandomState(1234)
    gen1 = (rng.randn(3, 4) for _ in range(5))
    gen2 = (rng.randn(3, 4) for _ in range(5))
    with pytest.raises(AssertionError):
        _runs_are_the_same(gen1, gen2, use_assert=True)

    gen1 = (i for i in range(5))
    gen2 = (i for i in range(5))
    _runs_are_the_same(gen1, gen2, use_assert=True)

    gen1 = (i for i in range(5))
    gen2 = (i if i<4 else 7 for i in range(5))
    with pytest.raises(AssertionError):
        _runs_are_the_same(gen1, gen2, use_assert=True)


if __name__ == '__main__':
    test_variable_matches_between_runs()
    test_assert_variable_matches_between_runs()



import Queue
import os
import socket
import sys
import threading
import time
import uuid
import pickle
from collections import namedtuple
from artemis.general.should_be_builtins import is_lambda
from artemis.plotting.matplotlib_backend import get_plotting_server_address
from artemis.remote.child_processes import PythonChildProcess
from artemis.remote.nanny import Nanny
from artemis.remote.file_system import check_config_file
from artemis.remote.port_forwarding import forward_tunnel
from artemis.remote.utils import get_local_ips, send_size, recv_size, check_ssh_connection

_to_subprocess_queue = None
_id_queue = None
_nanny = None
comm_terminate_event = threading.Event()

DBPlotMessage = namedtuple('DBPlotMessage', ['plot_id', 'dbplot_args'])


def dbplot_remotely(arg_locals):
    """
    This method should be called from dbplot immedeatly, in case we should plot remotely.
    arg_locals con
    :param arg_locals: A dict of arguments with which dbplot was called.
    """
    global _to_subprocess_queue
    global _id_queue

    assert not is_lambda(arg_locals['plot_type']), "dbplot in server mode does not accept lambda. Use partial instead"

    if _to_subprocess_queue is None:
        # This is the first call to dbplot, we need to set up the server.
        set_up_plotting_server()

    # The data is being prepared and sent to the remote plotting server
    unique_plot_id = str(uuid.uuid4())
    data_to_send = DBPlotMessage(plot_id = unique_plot_id, dbplot_args=arg_locals)
    serialized_command = pickle.dumps(data_to_send, protocol=2)
    _to_subprocess_queue.put(serialized_command)

    # Now we wait (or not) for the remote plot to be rendered.
    # Todo: This does not work reliably at the moment when several concurrent threads use dbplot. One thread might have dequeued the plot_id that belongs the other thread, such that
    # the other thread will never receive the plot_id. In case the user specified a positive waiting duration, the second call will then not return earlier although the plot might
    # have been rendered already
    wait_for_display_sec = arg_locals["wait_for_display_sec"]
    if wait_for_display_sec != 0:
        if wait_for_display_sec <0:
            wait_for_display_sec = sys.maxint # Todo: There is no timeout here
        begin = time.time()
        time_left = wait_for_display_sec
        try:
            while _id_queue.get(timeout=time_left) != unique_plot_id:
                time_used = time.time() - begin
                time_left = wait_for_display_sec - time_used
                if time_left < 0:
                    break
        except Queue.Empty:
            pass

def deconstruct_plotting_server():
    global _nanny
    global _to_subprocess_queue
    global _id_queue

    if _nanny:
        _nanny.deconstruct()

    _to_subprocess_queue = None
    _id_queue = None
    _nanny = None

def set_up_plotting_server():
    """
    Sets up the plotting server.
    """

    print("Setting up Plotting Server")

    # First we generate the system call that starts the server
    # TODO: This assumes the same installation path relative to the home-dir on the local machine as on the remote machine
    file_to_execute = os.path.join(os.path.dirname(__file__), 'plotting_server.py')
    file_to_execute = file_to_execute.replace(os.path.expanduser("~"),"~",1)
    plotting_server_address = get_plotting_server_address()
    if plotting_server_address == "":
        plotting_server_address = "127.0.0.1"
    if plotting_server_address in get_local_ips():
        command = ["python", "-u", file_to_execute]
    else:
        check_config_file(plotting_server_address) # Make sure all things are set
        check_ssh_connection(plotting_server_address) # Make sure the SSH-connection works
        command =["export DISPLAY=:0.0;", "python","-u", file_to_execute]
        # TODO: Setting DISPLAY to :0.0 is a heuristic at the moment. I don't understand yet how these DISPLAY variables are set.

    # With the command set up, we can instantiate a child process and start it. Also we want to forward stdout and stderr from the remote process asynchronously.
    global _nanny
    _nanny = Nanny()
    cp = PythonChildProcess(ip_address=plotting_server_address, command=command, name="Plotting_Server",set_up_port_for_structured_back_communication=True)
    _nanny.register_child_process(cp,monitor_for_termination=False,monitor_if_stuck_timeout=None,)
    _nanny.execute_all_child_processes(blocking=False)
    back_comm_queue = cp.get_queue_from_cp()
    try:
        is_debug_mode = getattr(sys, 'gettrace', None)
        timeout = None if is_debug_mode() else 10
        server_message = back_comm_queue.get(block=True,timeout=timeout)
    except Queue.Empty:
        print("The Plotting Server did not respond for 10 seconds. It probably crashed")
        sys.exit(1)

    try:
        port = int(server_message.dbplot_message)
    except ValueError:
        print("There was an incorrect string on the remote server's stdout. Make sure the server first communicates a port number. Received:\n {}".format(server_message.dbplot_message))
        sys.exit(0)

    # In the remote setting we don't want to rely on the user correctly specifying their firewalls. Therefore we need to set up port forwarding through ssh:
    # Also, we have the ssh session open already, so why not reuse it.
    if plotting_server_address not in get_local_ips():
        ssh_conn = cp.get_ssh_connection()
        # Needs to be in a thread since the call blocks forever.
        # Todo: this ssh tunnel is opened system-wide. That means that any subsequent attempts to open the ssh-tunnel (by another dbplot-using process, for example)
        # will perform wiredly. As far as I have tested, nothing happenes and the port forwarfding works just fine in the second process, However when one of the two
        # processes terminates, the ssh-tunnel is closed for the other process as well.
        t3 = threading.Thread(target = forward_tunnel, kwargs={"local_port":port, "remote_host":plotting_server_address, "remote_port":port,"ssh_conn":ssh_conn})
        t3.setDaemon(True)
        t3.start()

    # Now attempt to connect to the plotting server
    server_address = ("localhost", port)
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    try:
        sock.connect(tuple(server_address))
    except:
        raise

    # Once connected, set up the asynchronous threads that forward every dbplot call to the server. We make this asynchronously for two reasons:
    # 1.) the queues involved can be shared between different threads (not processes), therefore allowing for asynchronous dbplot calls that are both correctly forwarded.
    # 2.) sending a plot away to the server now immediatly returns, independent on any socket-related communication delays that might exist.
    # (There shouldn't be any, but, you know, principle)
    global _to_subprocess_queue
    global _id_queue
    _to_subprocess_queue = Queue.Queue()
    _id_queue = Queue.Queue()
    t1 = threading.Thread(target=push_to_server, args=(_to_subprocess_queue, sock))
    t1.setDaemon(True)
    t1.start()
    # if blocking:
    t2 = threading.Thread(target=collect_from_server, args=(_id_queue, sock))
    t2.setDaemon(True)
    t2.start()


def push_to_server(queue, sock):
    while True:
        try:
            message = queue.get_nowait()
            send_size(sock, message)
        except Queue.Empty:
            time.sleep(0.01)
    sock.close()


def collect_from_server(queue, sock):
    while True:
        recv_message = recv_size(sock)
        queue.put(recv_message)
    sock.close()


from artemis.experiments.decorators import experiment_function, experiment_root, ExperimentFunction
from artemis.experiments.experiments import capture_created_experiments, load_experiment
from artemis.experiments.experiment_record import load_experiment_record



from numpy.random import permutation
from sklearn import svm, datasets
from artemis.experiments import experiment_function


@experiment_function
def demo_iris_svm(C=1.0, gamma=0.7):

    iris = datasets.load_iris()
    perm = permutation(iris.target.size)
    iris.data = iris.data[perm]
    iris.target = iris.target[perm]
    clf = svm.SVC(C, 'rbf', gamma=gamma)
    clf.fit(iris.data[:90], iris.target[:90])
    return clf.score(iris.data[90:], iris.target[90:])


X=demo_iris_svm.add_variant(C=0.5)
X.add_variant(gamma=0.9)

if __name__ == '__main__':
    demo_iris_svm.browse()



from artemis.general.should_be_builtins import all_equal

__author__ = 'peter'
import numpy as np


def is_pareto_efficient_dumb(costs):
    """
    :param costs: An (n_points, n_costs) array
    :return: A (n_points, ) boolean array, indicating whether each point is Pareto efficient
    """
    is_efficient = np.ones(costs.shape[0], dtype = bool)
    for i, c in enumerate(costs):
        is_efficient[i] = np.all(np.any(costs>=c, axis=1))
    return is_efficient


def is_pareto_efficient(costs):
    """
    :param costs: An (n_points, n_costs) array
    :return: A (n_points, ) boolean array, indicating whether each point is Pareto efficient
    """
    is_efficient = np.ones(costs.shape[0], dtype = bool)
    for i, c in enumerate(costs):
        if is_efficient[i]:
            is_efficient[is_efficient] = np.any(costs[is_efficient]<=c, axis=1)  # Remove dominated points
    return is_efficient


def is_pareto_efficient_ixs(costs):

    candidates = np.arange(costs.shape[0])
    for i, c in enumerate(costs):
        if 0 < np.searchsorted(candidates, i) < len(candidates):  # If this element has not yet been eliminated
            candidates = candidates[np.any(costs[candidates]<=c, axis=1)]
    is_efficient = np.zeros(costs.shape[0], dtype = bool)
    is_efficient[candidates] = True
    return is_efficient


def find_pareto_ixs(cost_arrays):
    """
    :param cost_arrays: A collection of nd-arrays representing a grid of costs for different indices.
    :return: A tuple of indices which can be used to index the pareto-efficient points.
    """
    assert all_equal([c.shape for c in cost_arrays])
    flat_ixs, = np.nonzero(is_pareto_efficient(np.reshape(cost_arrays, (len(cost_arrays), -1)).T), )
    ixs = np.unravel_index(flat_ixs, dims=cost_arrays[0].shape)
    return ixs



import itertools

from six import string_types

from artemis.general.should_be_builtins import all_equal_deprecated, all_equal
from six.moves import xrange


def build_table(lookup_fcn, row_categories, column_categories, clear_repeated_headers = True, prettify_labels = True,
            row_header_labels = None, remove_unchanging_cols = False):
    """
    Build the rows of a table.  You can feed these rows into tabulate to generate pretty things.

        Example (requires installing tabulate (pip install tabulate):
        For the table of total utility in prisoner's dillema (see https://en.wikipedia.org/wiki/Prisoner%27s_dilemma):

        def lookup_function(prisoner_a_choice, prisoner_b_choice):
            total_utility = \
                2 if prisoner_a_choice=='cooperate' and prisoner_b_choice=='cooperate' else \
                3 if prisoner_a_choice != prisoner_b_choice else \
                4 if prisoner_b_choice=='betray' and prisoner_a_choice=='betray' \
                else bad_value((prisoner_a_choice, prisoner_b_choice))
            return total_utility

        rows = build_table(lookup_function, row_categories=['cooperate', 'betray'], column_categories=['cooperate', 'betray'])
        print tabulate.tabulate(rows)

        ---------  ---------  ------
                   Cooperate  Betray
        Cooperate  2          3
        Betray     3          4
        ---------  ---------  ------

        See more examples in test_tables.

    :param lookup_fcn: A function of the form:
        data = lookup_fcn(row_info, column_info)
        Where:
            row_info is a tuple of data identifying the row.
            col_info is a tuple of data identifying the column
    :param row_categories: A list<list<str>> of categories that will make up the rows
    :param column_categories: A list<list<str>> of catefories that will make up the columns
    :param clear_repeated_headers: True to not repeat row headers.
    :param row_header_labels: Labels for the row headers.
    :param remove_unchanging_cols: Remove columns for which all d
    :return: A list of rows.
    """
    # Now, build that table!
    single_row_category = all(isinstance(c, string_types) for c in row_categories)
    single_column_category = all(isinstance(c, string_types) for c in column_categories)

    if single_row_category:
        row_categories = [row_categories]
    if single_column_category:
        column_categories = [column_categories]
    if row_header_labels is not None:
        assert len(row_header_labels) == len(row_categories)
    rows = []
    column_headers = list(zip(*itertools.product(*column_categories)))
    for i, c in enumerate(column_headers):
        row_header = row_header_labels if row_header_labels is not None and i==len(column_headers)-1 else [' ']*len(row_categories)
        row = row_header+(blank_out_repeats(c) if clear_repeated_headers else list(c))
        rows.append([prettify_label(el) for el in row] if prettify_labels else row)
    last_row_data = [' ']*len(row_categories)
    for row_info in itertools.product(*row_categories):
        if clear_repeated_headers:
            row_header, last_row_data = zip(*[(h, h) if lh!=h else (' ', lh) for h, lh in zip(row_info, last_row_data)])
        else:
            row_header = row_info
        if prettify_labels:
            row_header = [prettify_label(str(el)) for el in row_header]
        data = [lookup_fcn(row_info[0] if single_row_category else row_info, column_info[0] if single_column_category else column_info) for column_info in itertools.product(*column_categories)]
        rows.append(list(row_header) + data)
    assert all_equal((len(r) for r in rows)), "All rows must have equal length.  They now have lengths: {}".format([len(r) for r in rows])

    if remove_unchanging_cols:
        for col_ix in range(len(rows[0]))[::-1]:
            if all_equal([row[col_ix] for row in rows[len(column_headers):]]):
                for row in rows:
                    del row[col_ix]
    return rows


def prettify_label(label):
    return (label[0].upper() + label[1:]).replace('_', ' ')


def blank_out_repeats(sequence, replace_with=' '):

    new_sequence = list(sequence)
    for i in xrange(len(new_sequence)-1, 0, -1):
        if new_sequence[i]==new_sequence[i-1]:
            new_sequence[i] = replace_with
    return new_sequence



from artemis.config import get_artemis_config_value
from matplotlib import pyplot as plt
__author__ = 'peter'

_plotting_mode = get_artemis_config_value(section='plotting', option='mode')

if _plotting_mode == 'safe':

    def redraw_figure(fig=None):
        plt.draw()
        plt.pause(0.00001)

elif _plotting_mode == 'fast':

    def redraw_figure(fig=None):
        if fig is None:
            fig = plt.gcf()
        fig.canvas.flush_events()
        plt.show(block=False)
        plt.show(block=False)

else:
    raise Exception("Unknown plotting mode: {}".format(_plotting_mode))



from artemis.fileman.local_dir import get_artemis_data_path
from artemis.plotting.saving_plots import save_figure, show_saved_figure
import matplotlib.pyplot as plt
import numpy as np
import pytest

__author__ = 'peter'


@pytest.mark.skipif(True, reason='Opens pdf with browser, which causes test server to hang.  Not a terribly important test anyway.')
def test_save_and_show_figure():

    fig = plt.figure()
    plt.imshow(np.random.randn(10, 10))
    plt.title('Test Figure')
    path = get_artemis_data_path('tests/test_fig.pdf')
    save_figure(fig, path = path)
    show_saved_figure(path)


@pytest.mark.skipif(True, reason='Opens pdf with browser, which causes test server to hang.  Not a terribly important test anyway.')
def test_save_and_show_figure_2():

    fig = plt.figure()
    plt.imshow(np.random.randn(10, 10))
    plt.title('Test Figure')
    path = get_artemis_data_path('tests/test_fig')
    path = save_figure(fig, path = path)
    show_saved_figure(path)


@pytest.mark.skipif(True, reason='Opens pdf with browser, which causes test server to hang.  Not a terribly important test anyway.')
def test_save_and_show_figure_3():

    fig = plt.figure()
    plt.imshow(np.random.randn(10, 10))
    plt.title('Test Figure')
    path = get_artemis_data_path('tests/test_fig.with.strangely.formatted.ending')
    path = save_figure(fig, path = path, ext='pdf')
    show_saved_figure(path)


if __name__ == '__main__':
    test_save_and_show_figure()



import atexit
import signal
import sys
import threading
import time

class ManagedChildProcess(object):
    def __init__(self,cp,monitor_for_termination,monitor_if_stuck_timeout):
        self.cp = cp
        self.monitor_for_termination = monitor_for_termination
        self.monitor_if_stuck_timeout = monitor_if_stuck_timeout

    def get_process(self):
        return self.cp

    @property
    def name(self):
        return self.cp.name

    def monitor_if_stuck(self):
        return self.monitor_if_stuck_timeout is not None

    def kill(self, signal=signal.SIGINT):
        return self.cp.kill(signal)

    def is_alive(self):
        return self.cp.is_alive()

    def get_name(self):
        return self.cp.get_name()

    def get_ip(self):
        return self.cp.get_ip()

    def get_id(self):
        return self.cp.get_id()

    def execute(self):
        return self.cp.execute_child_process()

    def get_id(self):
        return self.cp.get_id()

    def deconstruct(self, signum=signal.SIGINT, system_signal=False):
        return self.cp.deconstruct(signum)

class Nanny(object):
    '''
    Manages child processes. This class manages the start, live and deconstruction of child processes across different machines.
    '''
    def __init__(self,name=""):
        self.managed_child_processes = {}
        self.stdout_threads = {}
        self.name = name
        atexit.register(self.deconstruct)

    def register_child_process(self, cp, monitor_for_termination=True, monitor_if_stuck_timeout=None):
        '''
        This adds a child process to the Nanny.
        :param: monitor_for_termination: if set, this process' termination will trigger the shutdown process. If not set, this process' termination will go unnoticed
        :param: monitor_if_stuck_timeout: If set to a positive number, this process' communication is monitored. If not String has been passed over the Stdout channel of the child process in
         monitor_if_stuck_timeout seconds, the process is assumed to be stuck and the shutdown procedure is started.
        :return:
        '''
        assert monitor_if_stuck_timeout is None or (monitor_if_stuck_timeout > 0 and type(monitor_if_stuck_timeout) == int), "Please set monitor_if_stuck_timeout to a positive integer"

        self.managed_child_processes[cp.get_id()] = ManagedChildProcess(cp, monitor_for_termination,monitor_if_stuck_timeout)

    def get_child_processes(self):
        return {id:mcp.get_process() for id,mcp in self.managed_child_processes.items()}

    def execute_all_child_processes(self, time_out=1, stdout_stopping_criterium=lambda line:False, stderr_stopping_criterium =lambda line:False, blocking=True):
        '''
        Executes all child processes and starts managing communications. This method returns only when all child processes terminated.
        It might be the case that some child-processes hang or don't terminate. In this case, when one (the first) process
          terminates, all other processes are killed after time_out seconds. This behaviour is triggered also when

        :param: time_out: Grace Period for child processes to terminate before being killed
        :param: stdout_stopping_criterium: Receives the line from stdout. When evaluates to True, the stdout pipe is flushed and the termination request is set
        :param: stderr_stopping_criterium: Receives the line from stderr. When evaluates to True, the stdout pipe is flushed and the termination request is set
        :return:
        '''

        if blocking == False:
            t0 = threading.Thread(target=self.execute_all_child_processes,args=(time_out,stdout_stopping_criterium, stderr_stopping_criterium,True))
            t0.setDaemon(True)
            t0.start()
            all_cp_started = False
            while not all_cp_started:
                for cp in self.managed_child_processes.values():
                    if not cp.get_process().cp_started:
                        all_cp_started = False
                        break
                    else:
                        all_cp_started = True
                time.sleep(0.1)
            return

        termination_request_event = threading.Event()
        stdout_threads = {}
        stderr_threads = {}

        name_max_lenght = max([len(mcp.name) for mcp in self.managed_child_processes.values()])

        for i, id in enumerate(self.managed_child_processes.keys()):
            mcp =self.managed_child_processes[id]
            stdin, stdout, stderr = mcp.execute()

            prefix = mcp.name.ljust(name_max_lenght)+": "

            # True if in debug mode
            gettrace = getattr(sys, 'gettrace', None)
            timeout = mcp.monitor_if_stuck_timeout if not gettrace() else None # only set timeout if not in debug mode

            stdout_thread = threading.Thread(target=self._monitor_and_forward_child_communication,
                                             args=(stdout,sys.stdout,mcp.name,termination_request_event,stdout_stopping_criterium, prefix, timeout))

            stderr_thread = threading.Thread(target=self._monitor_and_forward_child_communication,
                                             args=(stderr,sys.stderr,mcp.name,termination_request_event,stderr_stopping_criterium, prefix, None))
            stdout_threads[mcp.get_id()] = stdout_thread
            stderr_threads[mcp.get_id()] = stderr_thread

            stdout_thread.start()
            stderr_thread.start()

        try:
            while not termination_request_event.wait(0.01):
                pass
        except KeyboardInterrupt:
            print("Nanny interrupted")
            self.deconstruct()
            sys.exit(1)

        # Grace period for other threads to shutdown
        time.sleep(time_out)
        for id,cp in self.managed_child_processes.items():
            if cp.is_alive():
                print(("Child Process %s at %s did not terminate %s seconds after the first process in cluster terminated. Terminating now." %(cp.get_name(), cp.get_ip(), time_out)))
                cp.deconstruct()
        for id,cp in self.managed_child_processes.items():
            if cp.is_alive():
                print(("Child Process %s at %s did not terminate. Force quitting now." %(cp.get_name(),cp.get_ip())))
                cp.deconstruct(signal.SIGKILL)


    def deconstruct(self):
        '''
        This method is called when SIGINT or SIGTERM are called.
        This aggressively deconstructs the Nanny and all child processes. Then, the signal is passed back to the original signal handlers.
        :return:
        '''
        for cp in self.managed_child_processes.values():
            cp.kill()
        time.sleep(1.0)
        for cp in self.managed_child_processes.values():
            if cp.is_alive():
                print(("Child Process %s at %s still alive, force terminating now"% (cp.name, cp.get_ip())))
                cp.kill(signal=signal.SIGTERM)


    def _monitor_and_forward_child_communication(self, source_pipe, target_pipe,process_name, termination_request_event=None, stopping_criterium=None, prefix="", timeout=None):
        '''
        thread to forward communication from source_pipe to target_pipe
        :param source_pipe:
        :param target_pipe:
        :param termination_request_event: Is set once the source_pipe has closed and this thread terminates, or when ths stopping_criterium has been met.
        :param stopping_criterium:
        :param prefix:
        :return:
        '''
        if timeout is not None:
            line_printed_event = threading.Event()
            line_printed_event.clear()
            t = threading.Thread(target=self._output_monitoring_timer_thread,args=(process_name,line_printed_event,termination_request_event,timeout))
            t.setDaemon(True)
            t.start()
        with source_pipe:
            for line in iter(source_pipe.readline, b''):
                target_pipe.write("%s%s"%(prefix,line))
                target_pipe.flush()
                if timeout is not None:
                    line_printed_event.set()
                if stopping_criterium is not None and stopping_criterium(line):
                    break
            if termination_request_event is not None:
                termination_request_event.set() # The input pipe closed, this thread terminates and we would like everybody to terminate

    def _output_monitoring_timer_thread(self, process_name, line_printed_event,termination_request_event, timeout=1800): # 5min
        timeout_wait_start = time.time()
        while time.time() - timeout_wait_start <= timeout:
            if line_printed_event.is_set():
                line_printed_event.clear()
                timeout_wait_start = time.time()
            if termination_request_event.is_set():
                return
            time.sleep(1.0)
        if termination_request_event.is_set():
            return

        print("Timeout occurred after %.1f min, process %s from Nanny %s stuck"%(timeout/60., process_name, self.name))
        termination_request_event.set()


import numpy as np
import pytest
from pytest import raises
from six import string_types
from six.moves import xrange

from artemis.general.nested_structures import (flatten_struct, get_meta_object, NestedType,
                                               seqstruct_to_structseq, structseq_to_seqstruct, nested_map,
                                               get_leaf_values)


def test_flatten_struct():

    a = {
        'a': [1, 2, 3],
        'b': {
            'c': 4,
            'd': {'e': 5, 6: [7, 8]},
            },
        'c': 1
        }
    b = dict(flatten_struct(a))
    assert b["['a'][1]"]==2
    assert b["['b']['d'][6][1]"]==8

    class A(object):

        def __init__(self, other_object, b):
            self._a = other_object
            self._b = b

    thing = A(A(3, 'arr'), {'x': np.random.randn(3, 2), 'y': (3, 4, 5)})

    flat_thing = dict(flatten_struct(thing))

    assert flat_thing['._a._b']=='arr'
    assert flat_thing["._b['y'][0]"]==3


_complex_obj = [1, 2, {'a':(3, 4), 'b':['hey', 'yeah']}, 'woo']
_meta_obj = [int, int, {'a':(int, int), 'b':[str, str]}, str]


def test_get_meta_object():
    assert get_meta_object(_complex_obj) == [int, int, {'a':(int, int), 'b':[str, str]}, str]


def test_nested_type():
    # Example from docstring
    assert NestedType.from_data([1, 2, {'a': (3, 4.), 'b': 'c'}]) == NestedType([int, int, {'a': (int, float), 'b': str}])

    # Other tests
    nested_type = NestedType.from_data(_complex_obj)
    assert nested_type.is_type_for(_complex_obj)
    _complex_obj_2 = [1, 2, {'a': (3, 4, 5), 'b': ['hey', 'yeah']}, 'woo']
    _complex_obj_3 = [1, 2, {'a': (3, 5), 'b': ['hey', 'yeah']}, 'woo']
    assert not nested_type.is_type_for(_complex_obj_2)
    assert nested_type.is_type_for(_complex_obj_3)
    assert nested_type == NestedType(_meta_obj)
    leaves = [1, 2, 3, 4, 'hey', 'yeah', 'woo']
    assert nested_type.get_leaves(_complex_obj) == leaves
    assert nested_type.expand_from_leaves(leaves) == _complex_obj

    extended_leaves = leaves + ['x']
    with raises(TypeError):
        nested_type.expand_from_leaves(extended_leaves)

    # Change the type
    swapped_leaves = list(leaves)
    swapped_leaves[0] = 'eh?'
    with raises(TypeError):
        nested_type.expand_from_leaves(swapped_leaves)


def test_seqstruct_to_structseq_and_inverse():

    a = [{'x': np.random.randn(2), 'y': [i, 'aaa']} for i in xrange(4)]
    b = seqstruct_to_structseq(a, as_arrays=True)

    assert b['x'].shape==(4, 2)
    assert np.array_equal(b['x'][2], a[2]['x'])
    assert np.array_equal(b['y'][0], np.arange(4))
    assert np.array_equal(b['y'][1], ['aaa']*4)

    c = structseq_to_seqstruct(b)
    assert np.array_equal(a[2]['x'], c[2]['x'])
    assert np.array_equal(a[3]['x'], c[3]['x'])
    assert np.array_equal(a[3]['y'][0], c[3]['y'][0])
    assert np.array_equal(a[3]['y'][1], c[3]['y'][1])


def test_nested_map():
    func = lambda x: x*2 if isinstance(x, (int, float)) else x+'  Not!' if isinstance(x, string_types) else x
    assert nested_map(func, 2)==4
    assert nested_map(func, 'God is dead.')=='God is dead.  Not!'
    assert nested_map(func, (1, 2, 3)) == (2, 4, 6)
    assert nested_map(func, [1, 2, None, {'a': 3, 'b': 'It works!'}]) == [2, 4, None, {'a': 6, 'b': 'It works!  Not!'}]
    with pytest.raises(AssertionError):
        assert nested_map(lambda a, b: a+b, {'a': 1, 'b': [2, 3]}, {'a': 4, 'XXX': [5, 6]}) == {'a': 5, 'b': [7, 9]}
    with pytest.raises(AssertionError):
        assert nested_map(lambda a, b: a+b, {'a': 1, 'b': [2, 3]}, {'a': 4, 'b': [5, 6], 'c': [7]}) == {'a': 5, 'b': [7, 9]}


def test_get_leaf_values():
    assert get_leaf_values([6]+[{'x': 3, 'y': [i, 'aaa']} for i in xrange(4)]) == [6, 3, 0, 'aaa', 3, 1, 'aaa', 3, 2, 'aaa', 3, 3, 'aaa']


def test_nested_map_with_container_func():

    data = {'a': [1, 2, 3], 'b': [3, 4, 5]}
    result = nested_map(lambda x: np.array(x), data, is_container_func=lambda x: isinstance(x, dict))
    assert isinstance(result['a'], np.ndarray)
    assert np.array_equal(result['a'], [1, 2, 3])
    assert isinstance(result['b'], np.ndarray)
    assert np.array_equal(result['b'], [3, 4, 5])


def test_none_bug():

    a = {'a': 1, 'b': None, 'c': [1, 2, None]}

    fa = flatten_struct(a, first_dict_is_namespace=True)
    assert dict(fa) == {'a': 1, 'b': None, 'c[0]': 1, 'c[1]': 2, 'c[2]': None}


if __name__ == '__main__':
    test_flatten_struct()
    test_get_meta_object()
    test_nested_type()
    test_seqstruct_to_structseq_and_inverse()
    test_nested_map()
    test_get_leaf_values()
    test_nested_map_with_container_func()
    test_none_bug()



from six.moves.configparser import NoSectionError, NoOptionError
from pytest import raises
from artemis.fileman.config_files import get_config_path, get_config_value, set_non_persistent_config_value
import os
__author__ = 'peter'


def test_get_config_value():

    config_path = get_config_path('.testconfigrc')

    if os.path.exists(config_path):
        os.remove(config_path)

    value = get_config_value(config_filename='.testconfigrc', section='opts', option='setting1', default_generator=lambda: 'somevalue', write_default=True)
    assert value == 'somevalue'

    value = get_config_value(config_filename='.testconfigrc', section='opts', option='setting1', default_generator=lambda: 'someothervalue', write_default=True)
    assert value == 'somevalue'

    value = get_config_value(config_filename='.testconfigrc', section='opts', option='setting2', default_generator=lambda: 'blah', write_default=True)
    assert value == 'blah'

    value = get_config_value(config_filename='.testconfigrc', section='opts', option='setting1')
    assert value == 'somevalue'

    value = get_config_value(config_filename='.testconfigrc', section='opts', option='setting2')
    assert value == 'blah'

    with raises(NoSectionError):
        _ = get_config_value(config_filename='.testconfigrc', section='schmopts', option='setting3')

    with raises(NoOptionError):
        _ = get_config_value(config_filename='.testconfigrc', section='opts', option='setting3')

    with raises(AssertionError):
        _ = get_config_value(config_filename='.testconfigXXXrc', section='opts', option='setting3')

    set_non_persistent_config_value(config_filename='.testconfigrc', section='opts', option='setting2',value="bob")
    value = get_config_value(config_filename='.testconfigrc', section='opts', option='setting2')
    assert value == 'bob'

    value = get_config_value(config_filename='.testconfigrc', section='opts', option='setting2', use_cashed_config=False)
    assert value == 'blah'

    value = get_config_value(config_filename='.testconfigrc', section='opts', option='setting2')
    assert value == 'bob'

    set_non_persistent_config_value(config_filename='.testconfigrc', section='schmapts', option='setting2', value="bob")
    with raises(NoOptionError):
        _ = get_config_value(config_filename='.testconfigrc', section='schmapts', option='setting3')

    with raises(NoSectionError):
        _ = get_config_value(config_filename='.testconfigrc', section='schmapts', option='setting2', use_cashed_config=False)

    value = get_config_value(config_filename='.testconfigrc', section='schmapts', option='setting2')
    assert value == 'bob'

    os.remove(config_path)


if __name__ == '__main__':
    test_get_config_value()



from itertools import count
from artemis.general.checkpoint_counter import CheckPointCounter, Checkpoints

__author__ = 'peter'


def test_checkpoint_counter():

    points = [0, 0.2, 0.4, 0.6, 0.9]

    cpc = CheckPointCounter(points)

    checkpoints = []
    for i in count(0):
        points_passed, done = cpc.check(i/100.)
        if points_passed:
            checkpoints.append(i)
            if done:
                break

    assert len(checkpoints)==len(points)
    assert checkpoints == [0, 20, 40, 60, 90]
    # Only because all points are spaced wider than progress increments.
    # More generally len(checkpoints) will be <= len(points)


def test_checkpoints():

    is_test = Checkpoints(('exp', 10, .1))

    assert [a for a in range(100) if is_test()]==[0, 10, 22, 37, 54, 74, 97]

if __name__ == '__main__':
    test_checkpoint_counter()
    test_checkpoints()


import os

import pytest
import shutil

from artemis.config import get_artemis_config_value
from artemis.fileman.config_files import get_config_value
from artemis.fileman.local_dir import get_artemis_data_path
from artemis.plotting.matplotlib_backend import get_plotting_server_address
from artemis.remote.file_system import rsync, simple_rsync, check_config_file
from artemis.remote.utils import get_local_ips

ip_address = get_plotting_server_address()
is_local = ip_address in get_local_ips()


@pytest.mark.skipif(is_local, reason ="No sense for local ip")
def test_rsync():
    options = ["-r"]
    username = get_artemis_config_value(section=ip_address, option="username")

    from_path = get_artemis_data_path(relative_path="tmp/tests/", make_local_dir=True)
    with open(os.path.join(from_path, "test1"), "wb"):
        pass
    with open(os.path.join(from_path, "test2"), "wb"):
        pass

    to_path = "%s@%s:/home/%s/temp/"%(username, ip_address, username)
    assert rsync(options, from_path, to_path)
    shutil.rmtree(from_path)


@pytest.mark.skipif(is_local, reason ="No sense for local ip")
def test_simple_rsync():
    from_path = get_artemis_data_path(relative_path="tmp/tests/", make_local_dir=True)
    with open(os.path.join(from_path, "test1"), "wb"):
        pass
    with open(os.path.join(from_path, "test2"), "wb"):
        pass
    remote_path = "~/PycharmProjects/Distributed-VI/"

    assert simple_rsync(local_path=from_path, remote_path=remote_path, ip_address=ip_address, verbose=True)
    shutil.rmtree(from_path)

if __name__ == "__main__":
    test_simple_rsync()
    # test_rsync()



from abc import abstractmethod

import numpy as np

from artemis.ml.tools.processors import OneHotEncoding


__author__ = 'peter'


class IPredictor(object):

    @abstractmethod
    def train(self, input_data, target_data):
        """
        :param input_data:
        :param target_data:
        :return:
        """

    @abstractmethod
    def predict(self, input_data):
        """
        :return: The output given the input data
        """


class CategoricalPredictor(IPredictor):
    """
    A wrapper that transforms a predictor that outputs a vector into
    a predictor that outputs an integer "category" label.
    """
    def __init__(self, predictor, n_categories = None, argmax_outputs = True):
        self._predictor = predictor
        self._n_categories = n_categories
        self._encoder = None if n_categories is None else OneHotEncoding(n_categories)
        self._argmax_outputs = argmax_outputs

    def train(self, input_data, target_data):
        if self._encoder is None:
            raise Exception('If you call train before predict, you must provide the number of categories.')
        new_target_data = self._encoder(target_data)
        return self._predictor.train(input_data, new_target_data)

    def predict(self, input_data):
        out = self._predictor.predict(input_data)
        if self._argmax_outputs:
            if self._encoder is None:
                assert out.ndim==2
                self._n_categories = out.shape[1]
                self._encoder = OneHotEncoding(self._n_categories)
            return np.argmax(out, axis = 1)
        else:
            return out
